from pipython import GCSDevice, pitools, GCS2Commands

class Stepper:
    """Represents an axis connected to a PI controller.

    This class provides methods for connecting to and controlling the axis, 
    including referencing the stage to a known position, moving the stage, 
    querying the position, and configuring the output trigger.

    Parameters
    ----------
    controller_ID : str
        The ID of the PI controller to connect to.
    axis_ID : str
        The ID of the axis to control.

    Attributes
    ----------
    pidevice : GCSDevice
        The underlying PI device object.
    controller_ID : str
        The ID of the PI controller this axis is connected to.
    axis_ID : str
        The ID of this axis.
    """

    def __init__(self, controller_id, axis_id):
        """Initializes the StepperChain class.

        Parameters:
        ----------
        controller_id (str): The ID of the controller.
        axis_id (int): The ID of the axis.
        """
        self.controller_id = controller_id
        self.pidevice = GCSDevice(devname=self.controller_id)
        self.axis_id = axis_id
    
    def usb_plugged_devices(self):
        """
        Returns a list with the devices plugged through USB.
        
        Returns:
        --------
        A list object with the indeces of the connected devices
        """
        return self.pidevice.EnumerateUSB(mask=self.controller_id)
    
    def connect_pidevice(self):
        """
        Finds the plugged devices, activates the I/O interface to select the device of interest and 
        eventually connects to the selected device.        
        """
        devices = self.usb_plugged_devices()
        if not devices:
            raise Exception("There are no plugged devices! Please connect at least one device.")

        selected_device = self.select_device(devices)
        self.startup_USB_device(selected_device)

    def select_device(self, devices):
        """
        Displays the list of devices and prompts the user to select one.
        
        Parameters
        ----------
        devices : list
            List object with the connected devices (as strings)
        
        Returns:
        --------
        A string object with the name of the pi device to be connected.
        """            
        print('Number ---- Device')
        for i, device in enumerate(devices):
            print(f'{i}      ----  {device}')

        item = int(input('Input the index of the device to connect: '))
        return devices[item]

    def startup_USB_device(self, device):
        """
        Connects to the specified device.
        
        Parameters
        ----------
        device : str
            A string object with the name of the pi device to be connected.
        """
        self.pidevice.ConnectUSB(device)
        pitools.startup(self.pidevice, self.axis_id)
        
    def move_stage_to_ref(self, refmode):
        """
        Moves the selected controller towards the reference position.
        
        Parameters
        ----------
        refmode : str
            String defining the referencing position.
        """
        if refmode == 'FNL':
            print("Moving stage towards negative edge...")
            self.pidevice.FNL()
        elif refmode == 'FPL':
            print("Moving stage towards positive edge...")
            self.pidevice.FPL()
        pitools.waitontarget(self.pidevice)
        print(f"Stage: {GCS2Commands.qCST(self.pidevice)['1']} successfully referenced.")
    
    def get_curr_pos(self):
        """
        Returns the current position of the axis
        
        Returns
        --------
        A float object with the current position of the stage

        """
        return self.pidevice.qPOS('1')

    def set_velocity(self,velocity):
        """ 
        Set the velocity of motion in the ROM of the controller
        
        Parameters
        ----------
        velocity : float
            Float defining the velocity of motion
        """
        self.pidevice.VEL('1',velocity)

    def set_acceleration(self,acceleration):
        """ 
        Set the acceleration of motion in the ROM of the controller
        
        Parameters
        ----------
        acceleration : float
            Float defining the acceleration of motion
        """
        self.pidevice.ACC('1',acceleration)
        
    def get_velocity(self):
        """ 
        Get and returns the velocity of the device
        
        Returns
        ----------
        velocity : float
            Float defining the velocity of motion
        """
        velocity = GCS2Commands.qVEL(self.pidevice)['1']
        return velocity

    def get_acceleration(self):
        """
        Gets and returns the acceleration of the device
               
        Returns
        --------
        A float object defining the acceleration of motion
        """
        acceleration = GCS2Commands.qACC(self.pidevice)['1']
        return acceleration

        
    def move_stage_to_target(self,target):
        """ 
        Moves the device to target position
        
        Parameters
        ----------
        target : float
            Float defining the target position
        """
        self.pidevice.MOV(self.pidevice.axes,target)
        pitools.waitontarget(self.pidevice)

    def enable_out_trigger(self, trigger_type):
        """Configures and activate the output trigger for a given axis.
        
        Parameters
        ----------
        trigger_type : int
            Type of trigger to be output (6 == in Motion, 1 = Line trigger).
        """
        self.pidevice.CTO(1, 2, 1)
        self.pidevice.CTO(1, 3, trigger_type)
        # enable trigger output with the configuration defined above
        self.pidevice.TRO(1, True)
        
    def disable_out_trigger(self, trigger_type):
        """Configures and disable the output trigger for a given axis.
        
        Parameters
        ----------
        trigger_type : int
            Type of trigger to be output (6 == in Motion, 1 = Line trigger).
        """
        self.pidevice.CTO(1, 2, 1)
        self.pidevice.CTO(1, 3, trigger_type)
        # disable trigger output with the configuration defined above
        self.pidevice.TRO(1, False)
        
    def close_connection(self):
        """
        Close the connection and reset the axis property
        """
        self.pidevice.CloseConnection()    


import json
import sys

MAX_SPEED = 10  # mm/s
MIN_POS = 0     # mm
MAX_POS = 102   # mm
MIN_STEP = 5e-4 # mm
MAX_STEP = 102  # mm
MAX_ACC = 20    # mm/s^2

def input_validator():
    """
    Validates the scan parameters extracted from a JSON file and terminates the program if any validation error occurs.
    
    Returns:
        dict: A dictionary containing the input parameters for the scanning process.
    """
    inpars = extract_scan_pars()
    scan_pars = inpars["scan_pars"]
    try:
        validate(scan_pars)
        return inpars
    except ValueError as e:
        print(e.args[0])
        print("Closing program ...")
        sys.exit()

def extract_scan_pars():
    """
    Extracts the scan parameters from a JSON file.

    Returns:
        dict: A dictionary containing the input parameters for the scanning process.
    """
    with open('../input/input_dicts.json') as openPars:
        inpars = json.load(openPars)
    return inpars

def validate(scan_pars):
    """
    Validates all the input parameters of scan_pars.

    Args:
        scan_pars (dict): A dictionary containing the input parameters for the scanning process.

    Raises:
        ValueError: If any of the input parameters is invalid.
    """
    validate_type(scan_pars["type"])
    validate_scan_edges(scan_pars["scan_edges"])
    validate_stepsize(scan_pars["scan_edges"], scan_pars["stepsize"])
    validate_velocity(scan_pars["velocity"])
    validate_acceleration(scan_pars["acceleration"])
    
def validate_type(type):
    """
    Validates the type of the scanning process.

    Args:
        type (str): Type of the scan, either "continuous" or "discrete".

    Raises:
        ValueError: If the type is not "continuous" or "discrete".
    """
    if not (type == "continuous" or type == "discrete"):
        raise ValueError("Invalid input: type must be either continuous or discrete.")
    
def validate_scan_edges(scan_edges):
    """
    Validates the scan edges.

    Args:
        scan_edges (list): A list containing the starting and final points of the scan.

    Raises:
        ValueError: If the scan edges are out of range [0, 102].
    """
    if (scan_edges[0] < MIN_POS) or (scan_edges[0] > MAX_POS):
        raise ValueError("Invalid input: first scan edge is out of range [0,102].")
    if (scan_edges[1] < MIN_POS) or (scan_edges[1] > MAX_POS):
        raise ValueError("Invalid input: second scan edge is out of range [0,102].")
        
def validate_stepsize(scan_edges, stepsize):
    """
    Validates the step size of the scanning process.

    Args:
        scan_edges (list): A list containing the starting and final points of the scan.
        stepsize (float): The step size between scan points.

    Raises:
        ValueError: If the step size is out of range [0.0005, 102], or the starting point plus step size are
                    out of range.
    """
    if (stepsize < MIN_STEP) or (stepsize >= MAX_STEP):
        raise ValueError("Invalid input: step size is out of range [0.0005,102].")
    if scan_edges[0] < scan_edges[1]:
        if stepsize + scan_edges[0] > MAX_STEP:
            raise ValueError("Invalid input: first scan edge plus step size is out of range.")
    elif scan_edges[0] > scan_edges[1]:
        if scan_edges[0] - stepsize < MIN_STEP:
            raise ValueError("Invalid input: first scan edge minus step size is out of range.")
    else:
        raise ValueError("Invalid input: the scan edges coincide.")
        
def validate_velocity(velocity):
    """
    Validates the velocity attribute to ensure that it is within the range [0, 10] mm/s.

    Args:
        velocity (float): The velocity of the scan motion in mm/s.

    Raises:
        ValueError: If the velocity value is out of range [0, 10] mm/s.
    """
    if (velocity < 0) or (velocity > MAX_SPEED):
        raise ValueError("Invalid input: velocity value is out of range [0, 10] mm/s.")
        
def validate_acceleration(acceleration):
    """
    Validates the acceleration attribute to ensure that it is within the range [0, 20] mm/s^2.

    Args:
        acceleration (float): The acceleration of the scan motion in mm/s^2.

    Raises:
        ValueError: If the acceleration value is out of range [0, 20] mm/s^2.
    """
    if (acceleration < 0) or (acceleration > MAX_ACC):
        raise ValueError("Invalid input: acceleration value is out of range [0, 20] mm/s^2.")


import numpy as np

def rows_columns_continuous(delta, stepsize):
    """
    Calculate the number of rows and columns for a continuous scan with a PI controller and Zurich lock-in.

    Parameters:
    -----------
    delta : float
        The distance between the scan edges.
    stepsize : float
        The step size for the motion of the stepper.

    Returns:
    --------
    N_rows : int
        The number of rows for the data acquisition.
    N_cols : int
        The number of columns for the data acquisition.
    """
    N_rows = 1
    N_cols = int(np.floor(delta / stepsize)) + 1
    return N_rows, N_cols

def rows_columns_discrete(delta, stepsize, sampl_freq):
    """
    Calculate the number of rows and columns for a discrete scan with a PI controller and Zurich lock-in.
    
    Parameters:
    -----------
    delta : float
        The distance between the scan edges.
    stepsize : float
        The step size for the motion of the stepper.
    sampl_freq : float
        The sampling frequency of the Zurich lock-in.

    Returns:
    --------
    N_rows : int
        The number of rows for the data acquisition.
    N_cols : int
        The number of columns for the data acquisition.
    """
    N_rows = int(np.floor(delta / stepsize)) + 1
    N_cols = int(np.floor(0.05 * sampl_freq))
    return N_rows, N_cols

def evaluate_daq_pars(scan_pars):
    """
    Process input data for a 1D scan to return the complete DAQ parameters that suit the scan parameters.

    Parameters:
    -----------
    scan_pars : dict
        A dictionary containing scan parameters, including:
            - stepsize : float
                The step size for motin of the stepper
            - acceleration : float
                The acceleration of the stepper.
            - velocity : float
                The velocity of the stepper.
            - sampling_freq : float
                The sampling frequency of the Zurich lock-in
            - type : str
                The type of scan: 'continuous' or 'discrete'.
            - scan_edges : list[float]
                The edges of the scan range.

    Returns:
    --------
    daq_pars : dict
        A dictionary containing data acquisition parameters, including:
            - daq_columns : int
                The number of columns for the data acquisition.
            - daq_rows : int
                The number of rows for the data acquisition.
            - duration : float
                The duration of the triggered data acquisition.
            - mode : str
                The data acquisition mode; 'Linear' for continuous scan, 'Exact (on-grid)' for discrete scan.
            - trigger_type : str
                The trigger type, always set to 'HW trigger'.
            - trigger_edge : str
                The trigger edge.
            - holdoff : float
                The holdoff time for the trigger.
    """
    stepsize = scan_pars["stepsize"]
    acc = scan_pars["acceleration"]
    vel = scan_pars["velocity"]
    sampl_freq = scan_pars["sampling_freq"]
    scan_type = scan_pars["type"]
    scan_edges = scan_pars["scan_edges"]

    delta = delta_calculator(scan_edges)

    if scan_type == "continuous":
        duration = duration_calculator(delta, vel, acc)
        N_rows, N_cols = rows_columns_continuous(delta, stepsize)
        mode = "Linear"
        edge = "positive"
    else:
        duration = 0.05  # 50 milliseconds
        N_rows, N_cols = rows_columns_discrete(delta, stepsize, sampl_freq)
        mode = "Exact (on-grid)"
        edge = "negative"

    daq_pars = {
        "daq_columns": N_cols,
        "daq_rows": N_rows,
        "duration": duration,
        "mode": mode,
        "trigger_type": "HW trigger",
        "trigger_edge": edge,
        "holdoff": duration * 0.95,
    }
    return daq_pars

def delta_calculator(scan_edges):
    """
    Calculate the absolute value of the distance between the scan edges.

    Parameters:
    -----------
    scan_edges : list[float]
        The edges of the scan range.

    Returns:
    --------
    delta : float
        Absolute value of the distance between scan edges.
    """
    return abs(scan_edges[1] - scan_edges[0])

def duration_calculator(delta, vel, acc):
    """
    Calculate the duration of the triggered data acquisition for a continuous scan.

    Parameters:
    -----------
    delta : float
        The distance between the scan edges.
    vel : float
        The velocity of the stepper.
    acc : float
        The acceleration of the stepper.

    Returns:
    --------
    duration : float
        The duration of the triggered data acquisition.
    """
    if np.sqrt(acc * delta) > vel:
        duration = vel / acc + delta / vel
    else:
        duration = 2 * np.sqrt(delta / acc)
    return duration


from pizurscan.Scanner import Scanner
from pizurscan.OutputProcessor import *
from pipython import pitools,GCS2Commands,GCSDevice
from pizurscan.PI_commands import Stepper
from pizurscan.InputProcessor import *
from pizurscan.InputValidator import *
import pytest as pytest
from math import isclose
import numpy as np 

class TestStepper:
    def setup_method(self, method):
        """Setup any state tied to the execution of the given method in a class.
        `setup_method` is invoked for every test method of a class.
        in the Setup, an instance of the stepper class is created and a PI_device is connected
        """
        self.stepper = Stepper('C-663', 'L-406.40SD00')
        dev_name = "PI C-663 Mercury-Step  SN 0020550099"
        self.stepper.startup_USB_device(dev_name)

    def teardown_method(self, method):
        """teardown any state that was previously setup with a setup_method
        call. In particular, it calls close_connection.
        """
        self.stepper.close_connection()
        self.stepper = None
        
    def test_initialization_controller_axis(self):
        """Tests that when a Stepper object is initialized through
        the __init__ method, the attributes controller_id and axis_id are
        properly assigned."""
        controller = 'C-663'
        axis = 'L-406.40SD00'
        # make id of controller and axis attributes
        assert self.stepper.controller_id == controller
        assert self.stepper.axis_id == axis
            
    def test_usb_return_notempty_list(self):
        """Test that when a Stepper object is passed to 
        usb_plugged_devices, the returned list is not empty."""
        self.stepper.close_connection()
        assert self.stepper.usb_plugged_devices() != []
        
    def test_select_device(self):
        """Test that when select device is called on a list of devices,
        the one selected with the input is correct."""
        random_devs = ["Device1", "Device2", "Device3"]
        device = self.stepper.select_device(devices = random_devs)
        assert device in random_devs
    
    def test_startup_USB_device(self): 
        """Tests that when the PI device with the known name is connected
        with ConnectUSB and startup, a connection is established."""    
        self.stepper.close_connection()
        self.stepper.connect_pidevice()
        assert self.stepper.pidevice.IsConnected()
        
    def test_connection_is_established(self):
        """
        Tests that when a Stepper object is passed to connect_pidevice,
        a connection to the controller is successfully established.
        """
        assert self.stepper.pidevice.IsConnected()
  
    def test_connected_axis(self):
        """
        Tests that when a Stepper object is connected, the connected axis 
        is one ('1'). C663 controller supports connection with only 
        axis at a time. 
        """
        assert pitools.getaxeslist(self.stepper.pidevice, axes=None) == ['1']
                
    def test_negative_reference(self):
        """Tests that when a connected Stepper object is referenced to the
        negative edge of the axis through move_stage_to_ref, the final position
        is indeed the negative edge (0)."""
        self.stepper.move_stage_to_ref('FNL')
        assert isclose(self.stepper.get_curr_pos()['1'],0,rel_tol=1e-8)

    def test_positive_reference(self):
        """Tests that when a connected Stepper object is referenced to the
        positive edge of the axis through move_stage_to_ref, the final position
        is indeed the positive edge (102)."""
        self.stepper.move_stage_to_ref('FPL')
        assert isclose(self.stepper.get_curr_pos()['1'],102,rel_tol=1e-8)
           
    def test_velocity_is_set(self):
        """Tests that when the velocity of a connected Stepper object is set through set_velocity, 
        the velocity stored in the controller's ROM is indeed the set one."""
        self.stepper.set_velocity(5)
        assert isclose(self.stepper.get_velocity(),5,rel_tol=1e-8)
    
    def test_acceleration_is_set(self):
        """Tests that when the acceleration of a connected Stepper object is set through set_acceleration, 
        the acceleration stored in the controller's ROM is indeed the set one."""
        self.stepper.set_acceleration(5)
        assert isclose(self.stepper.get_acceleration(),5,rel_tol=1e-8)
       
    def test_target_is_reached(self):
        """Tests that when the connected Stepper object is moved to a certain target through 
        move_stage_to_target, the position when the motion is completed is equal to the target."""
        self.stepper.move_stage_to_ref('FNL')
        self.stepper.move_stage_to_target(10)
        assert isclose(self.stepper.get_curr_pos()['1'],10,rel_tol=1e-8)
    
    def test_trigger_is_enabled(self):
        """Tests that when the trigger of the connected Stepper object is set through 
        configure out_trigger, the type of trigger stored in the ROM is the selected one."""
        self.stepper.enable_out_trigger(6)
        assert self.stepper.pidevice.qCTO(1, 3)[1][3] == '6'
    
    def test_trigger_is_disabled(self):
        """Tests that when the trigger of the connected Stepper object is set through 
        configure out_trigger, the type of trigger stored in the ROM is the selected one."""
        self.stepper.disable_out_trigger(6)
        assert not self.stepper.pidevice.qTRO(1)[1]

    def test_connection_is_closed(self):
        """Tests that the connection of the Stepper object with controller is indeed closed 
        with close_connection."""
        self.stepper.close_connection()
        assert not self.stepper.pidevice.IsConnected()

class TestInputProcessor:
    def setup_method(self, method):
        """Setup any state tied to the execution of the given method in a class.
        `setup_method` is invoked for every test method of a class.
        """
        self.scanPars = {		
			    "type": "discrete",
			    "scan_edges": [0,1],
                "stepsize" : 0.1,
			    "velocity" : 1,
		    	"acceleration" : 1,
			    "sampling_freq" : 100
		        }
    
    def teardown_method(self, method):
        """teardown any state that was previously setup with a setup_method
        call. 
        """
        self.scanPars = None

    def test_delta_calculation(self):
        """Tests that when the function delta_calculator is called, it correctly evaluates
        the delta value given the input scan_edges."""
        expected_delta = 1
        calculated_delta = delta_calculator(self.scanPars["scan_edges"])
        assert isclose(calculated_delta,expected_delta,rel_tol = 1e-8)
    
    def test_delta_traslation_invariant(self):
        """Tests that when the both the scan_edges values are shifted by a constant number, 
        the delta values before and after translation calculated with delta_calculator are the same."""
        # assumes that scan_edges[1] + 10 < 102
        delta_before_traslation = delta_calculator(self.scanPars["scan_edges"])
        scan_edges = self.scanPars["scan_edges"]
        scan_edges[0] += 10
        scan_edges[1] += 10
        delta_after_translation = delta_calculator(scan_edges)
        assert isclose(delta_before_traslation,delta_after_translation,rel_tol=1e-8)
        
    def test_duration_velocity_not_constant(self):
        """Tests that when a constant velocity of the stepper is never reached, the computation 
        of the duration of the motion through duration_calculator is corrected.          
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        vel, acc = self.scanPars["velocity"], self.scanPars["acceleration"]
        duration = duration_calculator(delta,vel,acc)
        print(duration)
        assert isclose(duration,2,rel_tol = 1e-8)
    
    def test_duration_velocity_constant(self):
        """Tests that when a constant velocity of the stepper is reached, the computation 
        of the duration of the motion through duration_calculator is corrected.          
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        vel, acc = self.scanPars["velocity"], 2
        duration = duration_calculator(delta,vel,acc)
        assert isclose(duration,1.5,rel_tol=1e-8)

    def test_rows_continous_values(self):
        """Tests that when the type of scan is set to continuous, the number of rows
        that are calculated with rows_columns_continuous is correct. 
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        stepsize = self.scanPars["stepsize"]
        N_rows , _ = rows_columns_continuous(delta,stepsize)
        assert N_rows == 1
        
    def test_cols_continuous_values(self):
        """Tests that when the type of scan is set to continuous, the number of columns
        that are calculated with rows_columns_continuous is correct. 
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        stepsize = self.scanPars["stepsize"]
        _ , N_cols = rows_columns_continuous(delta,stepsize)
        assert N_cols == 11
        
    def test_rows_discrete_values(self):
        """Tests that when the type of scan is set to discrete, the number of rows
        that are calculated with rows_columns_discrete is correct. 
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        stepsize = self.scanPars["stepsize"]
        sampl_freq = self.scanPars["sampling_freq"]
        N_rows, _ = rows_columns_discrete(delta,stepsize,sampl_freq)
        assert N_rows == 11
    
    def test_cols_discrete_values(self):
        """Tests that when the type of scan is set to discrete, the number of columns
        that are calculated with rows_columns_discrete is correct. 
        """
        delta = delta_calculator(self.scanPars["scan_edges"])
        stepsize = self.scanPars["stepsize"]
        sampl_freq = self.scanPars["sampling_freq"]
        _ , N_cols = rows_columns_discrete(delta,stepsize,sampl_freq)
        assert N_cols == 5
        
    def test_rows_cols_types(self):
        """Tests that when rows and columns are calculated in discrete and continuous mode through 
        rows_columns_continuous and rows_columns_discrete, respectively, the outputted numbers are integer."""
        delta = delta_calculator(self.scanPars["scan_edges"])
        stepsize = self.scanPars["stepsize"]
        sampl_freq = self.scanPars["sampling_freq"]
        N_rows_cont, N_cols_cont = rows_columns_continuous(delta,stepsize)
        N_rows_disc, N_cols_disc = rows_columns_discrete(delta,stepsize,sampl_freq)
        tup = (N_rows_cont,N_cols_cont,N_rows_disc,N_cols_disc)
        assert all(isinstance(i, int) for i in tup)
    
    def test_daq_pars_continuous(self):
        """ Tests that when the type of scan is set to continuous, the values of the dictionary "daq_pars" 
        calculated with the function evaluate_daq_pars are correct.
        """
        self.scanPars["type"] = "continuous"
        print(self.scanPars)
        daq_pars = evaluate_daq_pars(self.scanPars)
    
        assert daq_pars["daq_columns"] == 11
        assert daq_pars["daq_rows"] == 1
        assert daq_pars["duration"] == 2
        assert daq_pars["mode"] == "Linear"
        assert daq_pars["trigger_type"] == "HW trigger"
        assert daq_pars["trigger_edge"] == "positive"
        assert isclose(daq_pars["holdoff"],2*0.95,rel_tol = 1e-8)

    def test_daq_pars_discrete(self):
        """ Tests that when the type of scan is set to discrete, the values of the dictionary "daq_pars" 
        calculated with the function evaluate_daq_pars are correct.
        """
        daq_pars = evaluate_daq_pars(self.scanPars)
        duration = 0.05
        assert daq_pars["daq_columns"] == 5
        assert daq_pars["daq_rows"] == 11
        assert isclose(daq_pars["duration"],duration,rel_tol=1e-8)
        assert daq_pars["mode"] == "Exact (on-grid)"
        assert daq_pars["trigger_type"] == "HW trigger"
        assert daq_pars["trigger_edge"] == "negative"
        assert isclose(daq_pars["holdoff"],duration*0.95,rel_tol = 1e-8)

class TestScanner:
    def setup_method(self, method):
        """Setup any state tied to the execution of the given method in a class.
        `setup_method` is invoked for every test method of a class.
        """
        self.inPars = {
            "scan_pars" : {		
                "type": "discrete",
                "scan_edges": [0,1],
                "stepsize" : 0.1,
                "velocity" : 1,
                "acceleration" : 1,
                "sampling_freq" : 100
                },
            "pi" : 	{	
                "ID":"C-663",
                "stage_ID": "L-406.40SD00",
                "refmode": "FNL",
                "trig_type":6
                }
                }
        self.scanner = Scanner(self.inPars)

    
    def teardown_method(self, method):
        """teardown any state that was previously setup with a setup_method
        call.
        """
        self.scanner = None
        self.inPars = None
                
    def test_initialization_pars(self):
        """Tests that when an instance of scanner is created, the inizialitation of the 
        object's attributes through __init__ method is correct."""
        assert self.scanner.PI["ID"] == "C-663"
        assert self.scanner.PI["stage_ID"] == "L-406.40SD00"
        assert self.scanner.PI["refmode"] == "FNL"
        assert self.scanner.PI["trig_type"] == 6
        assert self.scanner.scan_pars["type"] == "discrete"
        assert isclose(self.scanner.scan_pars["scan_edges"][0],0,rel_tol = 1e-8)
        assert isclose(self.scanner.scan_pars["scan_edges"][1],1,rel_tol = 1e-8)
        assert isclose(self.scanner.scan_pars["stepsize"],0.1,rel_tol = 1e-8)
        
    def test_dimension_target_positions(self):
        """Tests that when an instance of scanner is created,the dimension of the 
        targets array is correct." 
        """
        N_targets = 11
        assert len(self.scanner.targets) == N_targets
        
    def test_values_target_positions(self):
        """Tests that when an instance of scanner is created, the first and the final 
        point of targets are correct."""
        assert isclose(self.scanner.targets[0],self.scanner.scan_edges[0],rel_tol = 1e-8)
        assert isclose(self.scanner.targets[-1],self.scanner.scan_edges[1],rel_tol = 1e-8)  # by induction is right because of linspace
    
    def test_stepper_is_instantiated(self):
        """Tests that when an instance of scanner is created, the instance of Stepper
        is properly instantiated throught the __init__ method."""
        assert isinstance(self.scanner.stepper,Stepper)
    
    def test_context_manager_enter(self):
        """Tests that when an instance of scanner is created and the context manager is opened, 
        the stepper object connects to the controller and gets referenced to negative edge."""
        with self.scanner as scan:
            assert scan.stepper.pidevice.gcsdevice.IsConnected()
            assert isclose(scan.stepper.get_curr_pos()['1'],0,rel_tol=1e-8)

    def test_context_manager_exit(self):
        """Tests that when an instance of scanner is created and the context manager is opened and closed, 
        the connection is also closed"""
        with self.scanner as scan:
            pass
        assert not self.scanner.stepper.pidevice.gcsdevice.IsConnected()

    def test_discrete_scan(self):
        """Tests that when an instance of Scan1D is created and execute_discrete_scan
           is performed, the covered positions are the targeted ones."""
        with self.scanner as scan:
            cur_pos = scan.execute_discrete_scan()
        assert all(abs(pos - target) <= 0.5e-3 for pos, target in zip(cur_pos, self.scanner.targets))

class TestOutputProcessor:
    def setup_method(self, method):
        """Setup any state tied to the execution of the given method in a class.
        `setup_method` is invoked for every test method of a class.
        """
        
        self.scan_pars = {		
                        "type": "discrete",
                        "scan_edges": [0.3,0.],
                        "stepsize" : 0.001,
                        "velocity" : 1,
                        "acceleration" : 1,
                        "sampling_freq" : 100
                        }
        self.daq_pars = {
                        "daq_columns" : 100,
                        "daq_rows" : 300,
                        "duration" : 0.05,
                        "mode" : "Exact (on-grid)",
                        "trigger type" : "HW trigger",
                        "trigger edge" : "negative",
                        "holdoff" : 0.05*(0.95),
                        }   
        self.filename =  "dev4910_demods_0_sample_r_avg_00000.csv"  
                    
    def teardown_method(self, method):
        """teardown any state that was previously setup with a setup_method
        call.
        """
        self.scan_pars = None
        self.daq_pars = None

    def test_raw_data_values(self):
        """Tests that when when the data outputted by the Zurich lock-in are read through get_raw_data,
        the returned NumPy has the same values of the original one.
        """
        # find raw data column directly, select only non nan values
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        raw_column = raw_data[:,2]
        raw_column_not_nan = raw_column[~np.isnan(raw_column)]
        # find raw data through get_raw_data, select only non nan values
        returned_column = get_raw_data(self.filename)
        returned_column_not_nan = returned_column[~np.isnan(returned_column)]
        assert all(abs(raw - returned)/raw <= 1e-8 for raw, returned in zip(raw_column_not_nan,returned_column_not_nan))
    
    def test_dimension_averaged_subintervals_is_rows(self):
        """Tests tat when raw data are read with get_raw_data and evaluate_averaged_data is called, 
        the dimension of the returned array (with averaged subintervals) is equal to the 
        number of rows"""
        returned_column = get_raw_data(self.filename)
        N_rows, N_cols = self.daq_pars["daq_rows"], self.daq_pars["daq_columns"]
        avg_returned_column = evaluate_averaged_data(N_rows,N_cols,returned_column)
        N_rows_avg = 300
        assert self.daq_pars["daq_rows"] == N_rows_avg
        
    def test_average_data(self):
        """Tests that when raw data are read by get_raw_data and evaluate_average_data
        is called, the average of the averaged subintervals (of N_cols dimensions) is equal to the
        average of the overall raw_data. This is justified by subintervals of equal dimension
        """
        # find raw data column directly, select only non nan values
        raw_data = np.genfromtxt(self.filename,skip_header = 1, delimiter = ";")
        raw_column = raw_data[:,2]
        raw_column_not_nan = raw_column[~np.isnan(raw_column)]
        avg_raw= np.mean(raw_column_not_nan)
        # find raw data through get_raw_data, select only non nan values
        returned_column = get_raw_data(self.filename)
        returned_column_not_nan = returned_column[~np.isnan(returned_column)]        
        # average subintervals of returned_column_not_nan
        N_rows, N_cols = self.daq_pars["daq_rows"], self.daq_pars["daq_columns"]
        avg_returned_column = evaluate_averaged_data(N_rows,N_cols,returned_column_not_nan)
        # average averaged subintervals of returned_column_not_nan
        avg_avg_returned = np.mean(avg_returned_column)
        assert isclose(avg_avg_returned,avg_raw,rel_tol = 1e-8)
        
    def test_dimension_target_positions(self):
        """Tests that when target positions are evaluated through evaluate_target_position,
        the dimension of the targets array is correct."""
        scanedges = self.scan_pars["scan_edges"]
        stepsize = self.scan_pars["stepsize"]
        targets = evaluate_target_positions(scanedges,stepsize)
        N_targets = 301
        assert N_targets == len(targets)
        
    def test_values_target_positions(self):
        """Tests that when target positions are evaluated through evaluate_target_position,
        the first and the last point of the target array is corrected"""
        scanedges = self.scan_pars["scan_edges"]
        stepsize = self.scan_pars["stepsize"]
        targets = evaluate_target_positions(scanedges,stepsize)
        assert isclose(scanedges[0],targets[0],rel_tol = 1e-8)
        assert isclose(scanedges[1],targets[-1],rel_tol = 1e-8)  # by induction is right because of linspace

    def test_save_data_file(self):
        """Tests that when processed data are saved through save_data_file,
        an output file named "cleaned_1D_data.txt" with the correct values is stored in the output folder.
        """
        # create fictitious targets 
        targets  = np.array([1,2,3,4,5],dtype=float)
        # create fictitious signal values
        avg_val = np.random.random(size = len(targets))
        # save file to "cleaned_1D_data.txt"
        save_data_file(targets,avg_val)
        # expected output
        expected_output = ''
        for i in range(len(targets)):
            expected_output += '\b'+str(targets[i])+', '+'{:.6f}'.format(avg_val[i])+'\n'
        # Read the file contents and compare with expected output
        with open("../output/cleaned_1D_data.txt", "rb") as f:
            assert f.read(),expected_output
            
        
class TestInputValidator:
    """ A class to test that when invalid input scan parameters are inserted, an exeption is raised and the code stops the execution. 
     Although a lower-level control is already present in the internal PI subroutines, it takes time to be activated. 
     Therefore, this additional entrance control find out invalid inputs without the need of connecting to the PI instrument. 
    """
    def test_invalid_type(self):
        """Tests that when an invalid value of 'type' is passed to the validate_type function,
        an exception (ValueError) is raised."""
        type = "invalid"
        with pytest.raises(ValueError):
            validate_type(type)

    def test_invalid_scan_edges(self):
        """ Tests that when an invalid value of 'scan_edges' is passed to the validate_scan_edges function,
        an exception (ValueError) is raised """
        scan_edges = [-10, 110]
        with pytest.raises(ValueError):
            validate_scan_edges(scan_edges)
    
    def test_invalid_stepsize(self):
        """ Tests that an invalid value of 'stepsize' is passed to the validate_stepsize function,
        an exception (ValueError) is raised """
        stepsize = 200
        scan_edges = [0,102]
        with pytest.raises(ValueError):
            validate_stepsize(scan_edges,stepsize)

    def test_invalid_velocity(self):
        """ Tests that when an invalid value of 'velocity' is passed to the validate_velocity function, 
        an exception (valueError) is raised """
        velocity = -5
        with pytest.raises(ValueError):
            validate_velocity(velocity)

    def test_invalid_acceleration(self):
        """ Tests that when an invalid value of 'acceleration' is passed to the validate_velocity function,
        an exception (ValueError) is raised """
        acceleration = 30
        with pytest.raises(ValueError):
            validate_acceleration(acceleration)

def test_valid_input(self):
        """Tests that when corrected values of scan parameters are provided, all the input validations 
        are successfull."""
        scanPars = {		
                "type": "discrete",
                "scan_edges": [0,1],
                "stepsize" : 0.1,
                "velocity" : 1,
                "acceleration" : 1,
                "sampling_freq" : 100
                }
        assert validate(scanPars)


import numpy as np

def get_raw_data(filename):
    """
    Read raw data from the input file and return the third column, which contains the values of the measured signal.

    Args:
    - filename (str): The name of the input file.

    Returns:
    - raw_data (ndarray): A NumPy array containing the third column of the input file.
    """
    raw_data = np.genfromtxt(filename, skip_header=1, delimiter=";")
    return raw_data[:, 2]

def evaluate_averaged_data(N_rows, N_cols, raw_data):
    """
    Average the input data in case of a 1D discrete scan. Raw data is divided into chunks of dimension N_cols, 
    over which an average is performed. The output avg_data is in one-to-one relation with the spatial coordinates.

    Args:
    - N_rows (int): The number of rows that was set in the daq.
    - N_cols (int): The number of columns that was set in the daq.
    - raw_data (ndarray): A NumPy array containing the raw data.

    Returns:
    - avg_data (ndarray): A NumPy array containing the averaged data.
    """
    avg_data = np.empty(N_rows)
    for row in range(N_rows):
        avg_data[row] = np.mean(raw_data[row * N_cols:(row + 1) * N_cols])
    return avg_data

def evaluate_target_positions(scan_edges, stepsize):
    """
    Evaluate the partition of the target points for a 1D scan.

    Args:
    - scan_edges (list): A list containing the scan range.
    - stepsize (float): The step size of the stepper.

    Returns:
    - targets (ndarray): A NumPy array containing the target positions.
    """
    # calculate target points
    N_points = int(abs(scan_edges[1] - scan_edges[0]) / stepsize) + 1
    return np.linspace(scan_edges[0], scan_edges[1], N_points, endpoint=True)

def save_data_file(targets, avg_data):         
    """
    Save the cleaned 1D data to a file named "cleaned_1D_data.txt" in output folder.

    Args:
    - targets (ndarray): A NumPy array containing the target positions.
    - avg_data (ndarray): A NumPy array containing the averaged data.
    """
    out_name = "../output/cleaned_1D_data.txt"
    out_file = np.column_stack((targets, avg_data))
    np.savetxt(out_name, out_file, fmt="%10.6f", delimiter=",")

def save_processed_data(filename, scan_pars, daq_pars):
    """
    Process the 1D data, averaging it if necessary (discrete scan), and save it to 
    a file named "cleaned_1D_data.txt" that is stored in the "output" folder.

    Args:
    - filename (str): The name of the input file.
    - scan_pars (dict): A dictionary containing scan parameters.
    - daq_pars (dict): A dictionary containing data acquisition parameters.
    """
    # extract input values
    filename = "../output/" + filename
    N_rows = daq_pars["daq_rows"]
    N_cols = daq_pars["daq_columns"]

    # get target positions
    scan_edges = scan_pars["scan_edges"]
    stepsize = scan_pars["stepsize"]
    targets = evaluate_target_positions(scan_edges, stepsize)  
    
    # get output data
    raw_data = get_raw_data(filename)
    if scan_pars["type"] == "discrete":
        out_data = evaluate_averaged_data(N_rows, N_cols, raw_data)
    else: 
        out_data = raw_data
    # save data
    save_data_file(targets, out_data)


from PI_commands import Stepper
import numpy as np

class Scanner:
    """
    A class for performing a 1D scan using a PI device.

    Attributes:
        PI (dict): Dictionary of PI device information.
        scan_pars (dict): Dictionary of scan parameters.
        scan_edges (list): List containing the two edges of the scan. Axis will move from the leftward to the rightward.
        stepsize (float): Step size of the scan.
        targets (numpy.ndarray): Array of target positions for the scan.
        stepper (Stepper): Stepper object for controlling the PI device.

    Methods:
        __init__(self, InPars):
            Initializes Scanner object with input parameters.
        __enter__(self):
            Context manager enter method.
        __exit__(self, exc_type, exc_value, traceback):
            Context manager exit method.
        __connect_stepper(self):
            Connects to the PI device through a user-interface I/O.
        __reference_stepper(self):
            Setup the 1D scan in four steps.
        evaluate_target_positions(self):
            Evaluates the partition of the target points for a 1D scan.
        setup_motion_stepper(self):
            Stores input velocity, acceleration, and trigger type in the ROM of the device.
        init_scan(self):
            Disables the trigger and moves to the first target of the scan.
        execute_discrete_scan(self):
            Executes the 1D discrete scan by moving the axis on all the target positions.
        execute_continuous_scan(self):
            Executes the continuous scan by moving the axis to the last position.
    """

    def __init__(self, InPars):
        """ Initializes Scanner object with input parameters.
        
        Parameters:
        ----------
        - InPars : dict
            a dictionary of input parameters regarding the scan features
        
        Attributes:
        ----------
        - PI : dict
            a dictionary containing the PI controller and axis id
        - scan_pars: dict
            a dictionary containing the scan parameters
        - scan_edges : list
            a list containing the two edges of the scan. Axis will move from the leftward to the rightward.
        - stepsize : float
            a float containing the step size of the scan
        - targets : numpy.array
            a numpy.array containing the targets positions of the scan
        - stepper: Stepper
            a stepper object that instantiate Stepper class.
        """
        self.PI = InPars["pi"]
        self.scan_pars = InPars["scan_pars"]
        self.scan_edges = self.scan_pars["scan_edges"]
        self.stepsize = self.scan_pars["stepsize"]
        self.targets = self.evaluate_target_positions()
        self.stepper = Stepper(self.PI["ID"], self.PI["stage_ID"])


    def __enter__(self):
        """
        Context manager enter method.
        Establishes the connection with the pidevice as soon as a context manager is opened
        and references the axis to either the positive or the negative edge.
        
        Returns:
        -------
            Scanner: Scanner object connected to the pidevice and referenced
        """
        self.__connect_stepper()  
        self.__reference_stepper()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Context manager exit method that closes the connection with the pidevice.

        Parameters:
        ----------
        - exc_type : type
            The type of the exception raised, if any. None if no exception occurred.
        - exc_value : Exception
            The exception instance raised, if any. None if no exception occurred.
        - traceback : traceback
            The traceback object related to the exception, if any. None if no exception occurred.
        """
        self.stepper.close_connection()  # Close the connection with the pidevice


    def __connect_stepper(self):
        """
        Connects to the PI device through a user-interface I/O.
        """
        self.stepper.connect_pidevice()

    def __reference_stepper(self):
        """
        Moves stepper to the required reference position at the maximum velocity
        and acceleration. 
        """
        # Set high default values to obtain quick referencing
        max_vel = 10    # mm/s
        max_acc = 20    # mm/s^2
        self.stepper.set_velocity(max_vel)
        self.stepper.set_acceleration(max_acc)
        self.stepper.move_stage_to_ref(self.PI["refmode"])
        
    def evaluate_target_positions(self):
        """
        Evaluates the partition of the target points for a 1D scan.

        Returns:
        --------
        numpy.ndarray: Array of target positions.
        """
        Npoints = int(abs(self.scan_edges[1] - self.scan_edges[0]) / self.stepsize) + 1
        return np.linspace(self.scan_edges[0], self.scan_edges[1], Npoints, endpoint=True)

    def setup_motion_stepper(self):
        """
        Stores input velocity, acceleration, and trigger type in the ROM of the device.
        """
        self.stepper.enable_out_trigger(trigger_type=self.PI["trig_type"])
        self.stepper.set_velocity(self.scan_pars["velocity"])
        self.stepper.set_acceleration(self.scan_pars["acceleration"])
        
    def init_scan(self):
        """
        Disables the trigger that was previously set and moves to the first target of the scan.
        """
        self.stepper.disable_out_trigger(trigger_type=self.PI["trig_type"])
        self.stepper.move_stage_to_target(self.targets[0])
        self.setup_motion_stepper()
        
    def execute_discrete_scan(self):
        """
        Executes the 1D discrete scan by moving the axis to all the target positions.

        Returns:
        --------
        scan_pos : List of scanned positions.
        """
        self.init_scan()
        scan_pos = []
        for target in self.targets:
            self.stepper.move_stage_to_target(target)        
            cur_pos = self.stepper.get_curr_pos()
            print(f"Position: {cur_pos['1']:.3f}")
        return scan_pos

    def execute_continuous_scan(self):
        """
        Executes the continuous scan by moving the axis to the last position.
        """
        self.init_scan()
        self.stepper.move_stage_to_target(self.targets[-1])
